(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('node-zookeeper-client')) :
	typeof define === 'function' && define.amd ? define(['exports', 'node-zookeeper-client'], factory) :
	(factory((global.nodeZkTreecache = global.nodeZkTreecache || {}),global.nodeZookeeperClient));
}(this, (function (exports,nodeZookeeperClient) { 'use strict';

/**
 * @module node-zk-treecache
 */
/**
 * @private
 */
class DefaultTreeCacheSelector {
    traverseChildren(_) {
        return true;
    }
    acceptChild(_) {
        return true;
    }
}

/**
 * The kind of event emitted by the curator.
 * Please note that not all of them are implemented yet.
 */
var CuratorEventType;
(function (CuratorEventType) {
    /**
     * Corresponds to {@link CuratorFramework#create()}
     */
    // CREATE,
    /**
     * Corresponds to {@link CuratorFramework#delete()}
     */
    // DELETE,
    /**
     * Corresponds to {@link CuratorFramework#checkExists()}
     */
    CuratorEventType[CuratorEventType["EXISTS"] = 0] = "EXISTS";
    /**
       * Corresponds to {@link CuratorFramework#getData()}
       */
    CuratorEventType[CuratorEventType["GET_DATA"] = 1] = "GET_DATA";
    /**
       * Corresponds to {@link CuratorFramework#setData()}
       */
    // SET_DATA,
    /**
       * Corresponds to {@link CuratorFramework#getChildren()}
       */
    CuratorEventType[CuratorEventType["CHILDREN"] = 2] = "CHILDREN";
    /**
     * Corresponds to {@link CuratorFramework#sync(String, Object)}
     */
    // SYNC,
    /**
     * Corresponds to {@link CuratorFramework#getACL()}
     */
    // GET_ACL,
    /**
     * Corresponds to {@link CuratorFramework#setACL()}
     */
    // SET_ACL,
    /**
     * Corresponds to {@link CuratorFramework#transaction()}
     */
    // TRANSACTION,
    /**
     * Corresponds to {@link CuratorFramework#getConfig()}
     */
    // GET_CONFIG,
    /**
     * Corresponds to {@link CuratorFramework#reconfig()}
     */
    // RECONFIG,
    /**
     * Corresponds to {@link Watchable#usingWatcher(Watcher)} or {@link Watchable#watched()}
     */
    // WATCHED,
    /**
     * Corresponds to {@link CuratorFramework#watches()} ()}
     */
    // REMOVE_WATCHES,
    /**
     * Event sent when client is being closed
     */
    // CLOSING
})(CuratorEventType || (CuratorEventType = {}));
var ZooKeeperCode;
(function (ZooKeeperCode) {
    // noinspection JSUnusedGlobalSymbols
    ZooKeeperCode[ZooKeeperCode["OK"] = 0] = "OK";
    ZooKeeperCode[ZooKeeperCode["SYSTEMERROR"] = -1] = "SYSTEMERROR";
    ZooKeeperCode[ZooKeeperCode["RUNTIMEINCONSISTENCY"] = -2] = "RUNTIMEINCONSISTENCY";
    ZooKeeperCode[ZooKeeperCode["DATAINCONSISTENCY"] = -3] = "DATAINCONSISTENCY";
    ZooKeeperCode[ZooKeeperCode["CONNECTIONLOSS"] = -4] = "CONNECTIONLOSS";
    ZooKeeperCode[ZooKeeperCode["MARSHALLINGERROR"] = -5] = "MARSHALLINGERROR";
    ZooKeeperCode[ZooKeeperCode["UNIMPLEMENTED"] = -6] = "UNIMPLEMENTED";
    ZooKeeperCode[ZooKeeperCode["OPERATIONTIMEOUT"] = -7] = "OPERATIONTIMEOUT";
    ZooKeeperCode[ZooKeeperCode["BADARGUMENTS"] = -8] = "BADARGUMENTS";
    ZooKeeperCode[ZooKeeperCode["NEWCONFIGNOQUORUM"] = -13] = "NEWCONFIGNOQUORUM";
    ZooKeeperCode[ZooKeeperCode["RECONFIGINPROGRESS"] = -14] = "RECONFIGINPROGRESS";
    ZooKeeperCode[ZooKeeperCode["UNKNOWNSESSION"] = -12] = "UNKNOWNSESSION";
    ZooKeeperCode[ZooKeeperCode["APIERROR"] = -100] = "APIERROR";
    ZooKeeperCode[ZooKeeperCode["NONODE"] = -101] = "NONODE";
    ZooKeeperCode[ZooKeeperCode["NOAUTH"] = -102] = "NOAUTH";
    ZooKeeperCode[ZooKeeperCode["BADVERSION"] = -103] = "BADVERSION";
    ZooKeeperCode[ZooKeeperCode["NOCHILDRENFOREPHEMERALS"] = -108] = "NOCHILDRENFOREPHEMERALS";
    ZooKeeperCode[ZooKeeperCode["NODEEXISTS"] = -110] = "NODEEXISTS";
    ZooKeeperCode[ZooKeeperCode["NOTEMPTY"] = -111] = "NOTEMPTY";
    ZooKeeperCode[ZooKeeperCode["SESSIONEXPIRED"] = -112] = "SESSIONEXPIRED";
    ZooKeeperCode[ZooKeeperCode["INVALIDCALLBACK"] = -113] = "INVALIDCALLBACK";
    ZooKeeperCode[ZooKeeperCode["INVALIDACL"] = -114] = "INVALIDACL";
    ZooKeeperCode[ZooKeeperCode["AUTHFAILED"] = -115] = "AUTHFAILED";
    ZooKeeperCode[ZooKeeperCode["SESSIONMOVED"] = -118] = "SESSIONMOVED";
    ZooKeeperCode[ZooKeeperCode["NOTREADONLY"] = -119] = "NOTREADONLY";
    ZooKeeperCode[ZooKeeperCode["EPHEMERALONLOCALSESSION"] = -120] = "EPHEMERALONLOCALSESSION";
    ZooKeeperCode[ZooKeeperCode["NOWATCHER"] = -121] = "NOWATCHER";
    ZooKeeperCode[ZooKeeperCode["RECONFIGDISABLED"] = -123] = "RECONFIGDISABLED";
})(ZooKeeperCode || (ZooKeeperCode = {}));
/**
 *
 * @private
 * @param path
 * @param stat
 * @param data
 * @param returnCode
 */
function createGetDataCuratorEvent(path, stat, data, returnCode) {
    return {
        type: CuratorEventType.GET_DATA,
        path,
        stat: stat,
        data,
        returnCode
    };
}
/**
 *
 * @private
 * @param path
 * @param stat
 * @param children
 * @param returnCode
 */
function createGetChildrenCuratorEvent(path, stat, children, returnCode) {
    return {
        type: CuratorEventType.CHILDREN,
        path,
        stat: stat,
        children,
        returnCode
    };
}
function createExistsCuratorEvent(path, stat, returnCode) {
    return {
        type: CuratorEventType.EXISTS,
        path,
        stat: stat,
        returnCode
    };
}

/**
 * @module node-zk-treecache
 */
/**  */
var ZooKeeperEventType;
(function (ZooKeeperEventType) {
    ZooKeeperEventType[ZooKeeperEventType["NODE_CREATED"] = 0] = "NODE_CREATED";
    ZooKeeperEventType[ZooKeeperEventType["NODE_DELETED"] = 1] = "NODE_DELETED";
    ZooKeeperEventType[ZooKeeperEventType["NODE_DATA_CHANGED"] = 2] = "NODE_DATA_CHANGED";
    ZooKeeperEventType[ZooKeeperEventType["NODE_CHILDREN_CHANGED"] = 3] = "NODE_CHILDREN_CHANGED";
})(ZooKeeperEventType || (ZooKeeperEventType = {}));
function normalizeZooKeeperEvent(e) {
    const name = e.getName();
    const path = e.getPath();
    switch (Number(e.getType())) {
        case nodeZookeeperClient.Event.NODE_CREATED:
            return { type: ZooKeeperEventType.NODE_CREATED, name, path };
        case nodeZookeeperClient.Event.NODE_DELETED:
            return { type: ZooKeeperEventType.NODE_DELETED, name, path };
        case nodeZookeeperClient.Event.NODE_DATA_CHANGED:
            return { type: ZooKeeperEventType.NODE_DATA_CHANGED, name, path };
        case nodeZookeeperClient.Event.NODE_CHILDREN_CHANGED:
            return { type: ZooKeeperEventType.NODE_CHILDREN_CHANGED, name, path };
        default:
            throw new Error(`Unknown ZooKeeper event type received: ${e}`);
    }
}
function createWatcher(w) {
    return e => w.onZooKeeperEvent(normalizeZooKeeperEvent(e));
}
class AbstractGetBuilder {
    constructor() {
        this.watcher = undefined;
        this.callback = undefined;
    }
    withWatcher(watcher) {
        this.watcher = watcher;
        return this;
    }
    withCallback(cb) {
        this.callback = cb;
        return this;
    }
}
/**
 * @private
 * @param {CuratorFrameworkImpl} client
 * @returns {GetChildrenBuilder}
 */
function newGetChildrenBuilder(client) {
    return new GetChildrenBuilderImpl(client);
}
class GetChildrenBuilderImpl extends AbstractGetBuilder {
    constructor(client) {
        super();
        this.client = client;
    }
    forPath(path) {
        return new Promise((resolve, reject) => {
            const cb = (error, children, stat) => {
                if (!(error === undefined || error instanceof nodeZookeeperClient.Exception || error === null)) {
                    return reject(error);
                }
                const event = createGetChildrenCuratorEvent(path, stat, children, (error && error.code) || 0);
                if (this.callback)
                    this.callback.onCuratorEvent(this.client, event);
                if (children)
                    return resolve(children);
                else
                    return reject(error);
            };
            if (this.watcher !== undefined) {
                this.client.zkClient.getChildren(path, createWatcher(this.watcher), cb);
            }
            else {
                this.client.zkClient.getChildren(path, cb);
            }
        });
    }
}
function newGetDataBuilder(client) {
    return new GetDataBuilderImpl(client);
}
class GetDataBuilderImpl extends AbstractGetBuilder {
    constructor(client) {
        super();
        this.client = client;
    }
    forPath(path) {
        return new Promise((resolve, reject) => {
            const cb = (error, data, stat) => {
                if (!(error === undefined || error instanceof nodeZookeeperClient.Exception || error === null)) {
                    return reject(error);
                }
                const event = createGetDataCuratorEvent(path, stat, data, (error && error.code) || 0);
                if (this.callback)
                    this.callback.onCuratorEvent(this.client, event);
                if (error)
                    return reject(error);
                else
                    return resolve(data);
            };
            if (this.watcher !== undefined) {
                this.client.zkClient.getData(path, createWatcher(this.watcher), cb);
            }
            else {
                this.client.zkClient.getData(path, cb);
            }
        });
    }
}
function newExistsBuilder(client) {
    return new ExistsBuilderImpl(client);
}
class ExistsBuilderImpl extends AbstractGetBuilder {
    constructor(client) {
        super();
        this.client = client;
    }
    forPath(path) {
        return new Promise((resolve, reject) => {
            const cb = (error, stat) => {
                if (!(error === undefined || error === null || error instanceof nodeZookeeperClient.Exception)) {
                    return reject(error);
                }
                const rc = stat ? 0 : ZooKeeperCode.NONODE;
                const event = createExistsCuratorEvent(path, stat, (error && error.code) || rc);
                if (this.callback)
                    this.callback.onCuratorEvent(this.client, event);
                const ret = !!stat;
                if (error)
                    return reject(error);
                else
                    return resolve(ret);
            };
            if (this.watcher !== undefined) {
                this.client.zkClient.exists(path, createWatcher(this.watcher), cb);
            }
            else {
                this.client.zkClient.exists(path, cb);
            }
        });
    }
}

/**
 * TODO Desc module (in ConnectionState)
 * @module node-zk-treecache
 */
/**  */
/**
 * Represents state changes in the connection to ZK
 */
var ConnectionState;
(function (ConnectionState) {
    /**
     * Sent for the first successful connection to the server. NOTE: You will only
     * get one of these messages for any CuratorFramework instance.
     */
    ConnectionState[ConnectionState["CONNECTED"] = 0] = "CONNECTED";
    /**
       * There has been a loss of connection. Leaders, locks, etc. should suspend
       * until the connection is re-established.
       *
       * We don't have access to this event with the current ZooKeeper client...
       */
    ConnectionState[ConnectionState["SUSPENDED"] = 1] = "SUSPENDED";
    /**
       * A suspended, lost, or read-only connection has been re-established
       */
    ConnectionState[ConnectionState["RECONNECTED"] = 2] = "RECONNECTED";
    /**
       * <p>
       *     Curator will set the LOST state when it believes that the ZooKeeper session
       *     has expired. ZooKeeper connections have a session. When the session expires, clients must take appropriate
       *     action. In Curator, this is complicated by the fact that Curator internally manages the ZooKeeper
       *     connection. Curator will set the LOST state when any of the following occurs:
       *     a) ZooKeeper returns a {@link Watcher.Event.KeeperState#Expired} or {@link KeeperException.Code#SESSIONEXPIRED};
       *     b) Curator closes the internally managed ZooKeeper instance; c) The session timeout
       *     elapses during a network partition.
       * </p>
       *
       * <p>
       *     NOTE: see {@link CuratorFrameworkFactory.Builder#connectionHandlingPolicy(ConnectionHandlingPolicy)} for an important note about a
       *     change in meaning to LOST since 3.0.0
       * </p>
       */
    ConnectionState[ConnectionState["LOST"] = 3] = "LOST";
    /**
       * The connection has gone into read-only mode. This can only happen if you pass true
       * for {@link CuratorFrameworkFactory.Builder#canBeReadOnly()}. See the ZooKeeper doc
       * regarding read only connections:
       * <a href="http://wiki.apache.org/hadoop/ZooKeeper/GSoCReadOnlyMode">http://wiki.apache.org/hadoop/ZooKeeper/GSoCReadOnlyMode</a>.
       * The connection will remain in read only mode until another state change is sent.
       */
    ConnectionState[ConnectionState["READ_ONLY"] = 4] = "READ_ONLY";
})(ConnectionState || (ConnectionState = {}));
/**
 * Check if this state indicates that Curator has a connection to ZooKeeper
 *
 * @private
 * @return True if connected, false otherwise
 */
function isConnected(cs) {
    switch (cs) {
        case ConnectionState.CONNECTED:
            return true;
        case ConnectionState.RECONNECTED:
            return true;
        case ConnectionState.READ_ONLY:
            return true;
        default:
            return false;
    }
}

/**
 * @module node-zk-treecache
 */
/**  */
/**
 * Validate the provided znode path string
 * @param path znode path string
 * @return The given path if it was valid, for fluent chaining
 * @throws Error if the path is invalid
 */
function validatePath(path) {
    if (path == null) {
        throw new Error("Path cannot be null");
    }
    if (path.length === 0) {
        throw new Error("Path length must be > 0");
    }
    if (path.charAt(0) !== "/") {
        throw new Error("Path must start with / character");
    }
    if (path.length === 1) {
        // done checking - it's the root
        return path;
    }
    if (path.charAt(path.length - 1) === "/") {
        throw new Error("Path must not end with / character");
    }
    let reason = null;
    let lastc = "/";
    let chars = path.split("");
    let c;
    for (let i = 1; i < chars.length; lastc = chars[i], i++) {
        c = chars[i];
        if (c.charCodeAt(0) === 0) {
            reason = "null character not allowed @" + i;
            break;
        }
        else if (c === "/" && lastc === "/") {
            reason = "empty node name specified @" + i;
            break;
        }
        else if (c === "." && lastc === ".") {
            if (chars[i - 2] === "/" &&
                (i + 1 === chars.length || chars[i + 1] === "/")) {
                reason = "relative paths not allowed @" + i;
                break;
            }
        }
        else if (c === ".") {
            if (chars[i - 1] === "/" &&
                (i + 1 === chars.length || chars[i + 1] === "/")) {
                reason = "relative paths not allowed @" + i;
                break;
            }
        }
        else if ((c > "\u0000" && c < "\u001f") ||
            (c > "\u007f" && c < "\u009F") ||
            (c > "\ud800" && c < "\uf8ff") ||
            (c > "\ufff0" && c < "\uffff")) {
            reason = "invalid character @" + i;
            break;
        }
    }
    if (reason != null) {
        throw new Error('Invalid path string "' + path + '" caused by ' + reason);
    }
    return path;
}

/**
 * @module node-zk-treecache
 */
/**  */
/**
 * @private
 */
function checkNotNull(ref, message) {
    if (ref === null) {
        throw new Error(message);
    }
    else {
        return ref;
    }
}

/**
 * @module node-zk-treecache
 */
/**  */
const PATH_SEPARATOR = "/";
const ANYONE_ID_UNSAFE = new nodeZookeeperClient.Id("world", "anyone");
const OPEN_ACL_UNSAFE = [new nodeZookeeperClient.ACL(31, ANYONE_ID_UNSAFE)];
const EMPTY_CALLBACK = () => {
    return;
};
const DEFAULT_ACL_PROVIDER = {
    getDefaultAcl() {
        return null;
    },
    getAclForPath() {
        return null;
    }
};
/**
 * @private
 * @param path the path to split
 */
function split(path) {
    validatePath(path);
    return path.split(PATH_SEPARATOR).filter(s => s.length > 0);
}
// Verify if we are running ZK 3.5+, in which case we can use the create mode container (code 4)
function getCreateMode(asContainers) {
    return nodeZookeeperClient.CreateMode.PERSISTENT;
}
/**
 * Make sure all the nodes in the path are created. NOTE: Unlike File.mkdirs(), Zookeeper doesn't distinguish
 * between directories and files. So, every node in the path is created. The data for each node is an empty blob
 *
 * @private
 * @param zkClient     the client
 * @param path         path to ensure
 * @param makeLastNode if true, all nodes are created. If false, only the parent nodes are created
 * @param aclProvider  if not null, the ACL provider to use when creating parent nodes
 * @param asContainers if true, nodes are created as {@link CreateMode#CONTAINER} (need ZK > 3.5)
 * @param cb           the callback to call after having created the path
 */
function mkdirs(zkClient, path, makeLastNode, aclProvider, asContainers, cb) {
    validatePath(path);
    let s = split(path);
    if (!makeLastNode) {
        s.pop();
    }
    path = "/" + s.join("/");
    const mode = getCreateMode(asContainers);
    aclProvider = aclProvider || DEFAULT_ACL_PROVIDER;
    const acl = aclProvider.getAclForPath(path) ||
        aclProvider.getDefaultAcl() ||
        OPEN_ACL_UNSAFE;
    zkClient.mkdirp(path, acl, mode, cb || EMPTY_CALLBACK);
}
/**
 * Given a full path, return the node name. i.e. "/one/two/three" will return "three"
 *
 * @private
 * @param path the path
 * @return the node
 */
function getNodeFromPath(path) {
    validatePath(path);
    const last = path.split(PATH_SEPARATOR).pop();
    if (last === undefined)
        throw new Error(`Error while validating ${path}, it should have been valid`);
    return last;
}
/**
 * Given a parent path and a child node, create a combined full path
 *
 * @param parent the parent
 * @param child  the child
 * @return full path
 */
function makePath(parent, child) {
    return [parent.replace(/\/+$/, ""), child.replace(/^\/+/, "")].join("/");
}

/**
 * @module node-zk-treecache
 */
/**  */
function compareChildData(cd1, cd2) {
    if (cd1 === cd2)
        return true;
    if (!cd1.data && cd2.data)
        return false;
    if (cd1.data && !cd2.data)
        return false;
    if (cd1.data && cd2.data && !cd1.data.equals(cd2.data))
        return false;
    if (cd1.path !== cd2.path)
        return false;
    if ((cd1.stat && !cd2.stat) || (!cd1.stat && cd2.stat))
        return false;
    if (!cd1.stat || !cd2.stat)
        return true;
    if (cd1.stat.czxid.toString("base64") !== cd2.stat.czxid.toString("base64"))
        return false;
    if (cd1.stat.mzxid.toString("base64") !== cd2.stat.mzxid.toString("base64"))
        return false;
    if (cd1.stat.ctime.toString("base64") !== cd2.stat.ctime.toString("base64"))
        return false;
    if (cd1.stat.mtime.toString("base64") !== cd2.stat.mtime.toString("base64"))
        return false;
    if (cd1.stat.version !== cd2.stat.version)
        return false;
    if (cd1.stat.cversion !== cd2.stat.cversion)
        return false;
    if (cd1.stat.aversion !== cd2.stat.aversion)
        return false;
    if (cd1.stat.ephemeralOwner.toString("base64") !==
        cd2.stat.ephemeralOwner.toString("base64"))
        return false;
    if (cd1.stat.dataLength !== cd2.stat.dataLength)
        return false;
    if (cd1.stat.numChildren !== cd2.stat.numChildren)
        return false;
    return cd1.stat.pzxid.toString("base64") === cd2.stat.pzxid.toString("base64");
}
var NodeState;
(function (NodeState) {
    NodeState[NodeState["PENDING"] = 0] = "PENDING";
    NodeState[NodeState["LIVE"] = 1] = "LIVE";
    NodeState[NodeState["DEAD"] = 2] = "DEAD";
})(NodeState || (NodeState = {}));
class TreeNode {
    constructor(_path, parent, treeCacheView, logger) {
        this._path = _path;
        this.parent = parent;
        this.treeCacheView = treeCacheView;
        this.logger = logger;
        this._children = new Map();
        this._nodeState = NodeState.PENDING;
        this.depth = parent === null ? 0 : parent.depth + 1;
    }
    get path() {
        return this._path;
    }
    get nodeState() {
        return this._nodeState;
    }
    get children() {
        return this._children;
    }
    get childData() {
        return this._childData;
    }
    wasCreated() {
        this.refresh();
    }
    wasDeleted() {
        const oldChildData = this._childData;
        this._childData = null;
        const childMap = this._children;
        this._children = null;
        if (childMap !== null) {
            const childCopy = Array.from(childMap.values());
            childMap.clear();
            for (let child of childCopy) {
                child.wasDeleted();
            }
        }
        if (this.treeCacheView.treeState() === TreeState.CLOSED) {
            return;
        }
        const oldState = this._nodeState;
        if (oldState === NodeState.LIVE) {
            this.treeCacheView.publishEventTypeWithData(exports.TreeCacheEventType.NODE_REMOVED, oldChildData);
        }
        if (this.parent == null) {
            // Root node; use an exist query to watch for existence.
            this.treeCacheView.client
                .checkExists()
                .withWatcher(this)
                .withCallback(this)
                .forPath(this._path)
                .catch(reason => {
                this.logger.debug(`[Node ${this.path}]`, "wasDeleted#checkExists failed with reason", reason);
            });
        }
        else {
            // Remove from parent if we're currently a child
            const parentChildMap = this.parent.children;
            if (parentChildMap != null) {
                parentChildMap.delete(getNodeFromPath(this.path));
            }
        }
    }
    wasReconnected() {
        this.refresh();
        if (this.children !== null) {
            for (let child of this.children.values()) {
                child.wasReconnected();
            }
        }
    }
    onCuratorEvent(client, event) {
        const loggedEvent = Object.assign({}, event, { stat: "hidden" });
        this.logger.debug("onCuratorEvent( event = ", loggedEvent, ")");
        switch (event.type) {
            case CuratorEventType.EXISTS:
                if (this.parent !== null)
                    throw new Error("unexpected EXISTS on non-root node");
                if (event.returnCode === ZooKeeperCode.OK) {
                    if (this._nodeState === NodeState.DEAD) {
                        this._nodeState = NodeState.PENDING;
                    }
                    this.wasCreated();
                }
                break;
            case CuratorEventType.CHILDREN: {
                if (event.returnCode === ZooKeeperCode.OK) {
                    if (event.stat === undefined) {
                        this.logger.error("TreeCache got an event CHILDREN with a status code OK but an undefined stat");
                        break;
                    }
                    if (event.children === undefined) {
                        this.logger.error("TreeCache got an event CHILDREN with a status code OK but an undefined children");
                        break;
                    }
                    const newStat = event.stat;
                    const oldChildData = this._childData;
                    if (oldChildData &&
                        oldChildData.stat &&
                        oldChildData.stat.mzxid.toString("base64") ===
                            newStat.mzxid.toString("base64")) {
                        // Only update stat if mzxid is same, otherwise we might obscure
                        // GET_DATA event updates.
                        const newChildData = {
                            path: oldChildData.path,
                            stat: newStat,
                            data: oldChildData.data
                        };
                        if (compareChildData(oldChildData, newChildData)) {
                            this._childData = newChildData;
                        }
                    }
                    if (event.children.length < 1)
                        break;
                    // Create map if not previously done
                    let childMap = this._children;
                    if (childMap === null) {
                        childMap = new Map();
                        if (this._children === null) {
                            this._children = childMap;
                        }
                        else {
                            childMap = this._children;
                        }
                    }
                    // Present new children in sorted order for test determinism
                    const newChildren = [];
                    for (const child of event.children) {
                        if (!childMap.has(child) &&
                            this.treeCacheView.selector.acceptChild(makePath(this.path, child))) {
                            newChildren.push(child);
                        }
                    }
                    newChildren.sort();
                    for (const child of newChildren) {
                        const fullPath = makePath(this.path, child);
                        const node = new TreeNode(fullPath, this, this.treeCacheView, this.logger);
                        if (!childMap.has(child)) {
                            childMap.set(child, node);
                            node.wasCreated();
                        }
                    }
                }
                else if (event.returnCode === ZooKeeperCode.NONODE) {
                    this.wasDeleted();
                }
                break;
            }
            case CuratorEventType.GET_DATA: {
                if (event.returnCode === ZooKeeperCode.OK) {
                    if (event.stat === undefined) {
                        this.logger.error("TreeCache got an event GET_DATA with a status code OK but an undefined stat");
                        break;
                    }
                    const newStat = event.stat;
                    const toPublish = {
                        path: event.path,
                        stat: newStat,
                        data: event.data
                    };
                    let oldChildData = this._childData;
                    if (this.treeCacheView.cacheData) {
                        this._childData = toPublish;
                    }
                    else {
                        this._childData = {
                            path: event.path,
                            stat: newStat,
                            data: undefined
                        };
                    }
                    let added;
                    if (this.parent === null) {
                        // We're the singleton root.
                        const nodeState = this._nodeState;
                        this._nodeState = NodeState.LIVE;
                        added = nodeState !== NodeState.LIVE;
                    }
                    else {
                        if (this._nodeState === NodeState.PENDING) {
                            this._nodeState = NodeState.LIVE;
                            added = true;
                        }
                        else {
                            added = false;
                            // Ordinary nodes are not allowed to transition from dead -> live;
                            // make sure this isn't a delayed response that came in after death.
                            if (this.nodeState !== NodeState.LIVE)
                                return;
                        }
                    }
                    if (added) {
                        this.treeCacheView.publishEventTypeWithData(exports.TreeCacheEventType.NODE_ADDED, toPublish);
                    }
                    else {
                        if (!oldChildData ||
                            !oldChildData.stat ||
                            oldChildData.stat.mzxid.toString("base64") !==
                                newStat.mzxid.toString("base64")) {
                            this.treeCacheView.publishEventTypeWithData(exports.TreeCacheEventType.NODE_UPDATED, toPublish);
                        }
                    }
                }
                else if (event.returnCode === ZooKeeperCode.NONODE) {
                    this.wasDeleted();
                }
                break;
            }
        }
        this.treeCacheView.outstandingOps--;
        if (this.treeCacheView.outstandingOps === 0) {
            if (this.treeCacheView.isInitialized === false) {
                this.treeCacheView.isInitialized = true;
                this.treeCacheView.publishEventType(exports.TreeCacheEventType.INITIALIZED);
            }
        }
    }
    onZooKeeperEvent(event) {
        this.logger.debug(`[Node ${this.path}] onZooKeeperEvent`, event);
        try {
            switch (event.type) {
                case ZooKeeperEventType.NODE_CREATED:
                    if (this.parent !== null)
                        throw new Error("Unexpected NODE_CREATED on non-root node");
                    this.wasCreated();
                    break;
                case ZooKeeperEventType.NODE_CHILDREN_CHANGED:
                    this.refreshChildren();
                    break;
                case ZooKeeperEventType.NODE_DATA_CHANGED:
                    this.refreshData();
                    break;
                case ZooKeeperEventType.NODE_DELETED:
                    this.wasDeleted();
                    break;
            }
        }
        catch (e) {
            this.treeCacheView.handleException(e);
        }
    }
    refresh() {
        if (this.depth < this.treeCacheView.maxDepth &&
            this.treeCacheView.selector.traverseChildren(this.path)) {
            this.treeCacheView.outstandingOps += 2;
            this.doRefreshData();
            this.doRefreshChildren();
        }
        else {
            this.refreshData();
        }
    }
    refreshData() {
        this.treeCacheView.outstandingOps++;
        this.doRefreshData();
    }
    doRefreshData() {
        if (this.treeCacheView.treeState() === TreeState.STARTED) {
            this.treeCacheView.client
                .getData()
                .withWatcher(this)
                .withCallback(this)
                .forPath(this.path)
                .catch(reason => {
                this.logger.debug(`[Node ${this.path}]`, "doRefreshData() failed with reason", reason);
            });
        }
    }
    refreshChildren() {
        if (this.depth < this.treeCacheView.maxDepth &&
            this.treeCacheView.selector.traverseChildren(this.path)) {
            this.treeCacheView.outstandingOps++;
            this.doRefreshChildren();
        }
    }
    doRefreshChildren() {
        if (this.treeCacheView.treeState() === TreeState.STARTED) {
            this.treeCacheView.client
                .getChildren()
                .withWatcher(this)
                .withCallback(this)
                .forPath(this.path)
                .catch(reason => {
                this.logger.debug(`[Node ${this.path}]`, "doRefreshChildren() failed with reason", reason);
            });
        }
    }
}
class TreeCacheEvent {
    constructor(_type, _data, _date) {
        this._type = _type;
        this._data = _data;
        this._date = _date;
    }
    get type() {
        return this._type;
    }
    get data() {
        return this._data;
    }
    get path() {
        return this._data !== null ? this._data.path : null;
    }
    get date() {
        return this._date;
    }
    toString() {
        let data;
        if (this._data) {
            data = JSON.stringify(Object.assign({}, this._data, {
                stat: this._data.stat ? "defined" : "undefined"
            }));
        }
        else {
            data = "undefined";
        }
        return `TreeCacheEvent { type = ${this._type.toString()}, path = ${this
            .path}, data = ${data} }`;
    }
}

(function (TreeCacheEventType) {
    /**
     * A node was added.
     */
    TreeCacheEventType[TreeCacheEventType["NODE_ADDED"] = 0] = "NODE_ADDED";
    /**
       * A node's data was changed
       */
    TreeCacheEventType[TreeCacheEventType["NODE_UPDATED"] = 1] = "NODE_UPDATED";
    /**
       * A node was removed from the tree
       */
    TreeCacheEventType[TreeCacheEventType["NODE_REMOVED"] = 2] = "NODE_REMOVED";
    /**
       * Called when the connection has changed to {@link ConnectionState#SUSPENDED}
       * <p>
       * This is exposed so that users of the class can be notified of issues that *might* affect normal operation.
       * The TreeCache is written such that listeners are not expected to do anything special on this
       * event, except for those people who want to cause some application-specific logic to fire when this occurs.
       * While the connection is down, the TreeCache will continue to have its state from before it lost
       * the connection and after the connection is restored, the TreeCache will emit normal child events
       * for all of the adds, deletes and updates that happened during the time that it was disconnected.
       * </p>
       */
    TreeCacheEventType[TreeCacheEventType["CONNECTION_SUSPENDED"] = 3] = "CONNECTION_SUSPENDED";
    /**
       * Called when the connection has changed to {@link ConnectionState#RECONNECTED}
       * <p>
       * This is exposed so that users of the class can be notified of issues that *might* affect normal operation.
       * The TreeCache is written such that listeners are not expected to do anything special on this
       * event, except for those people who want to cause some application-specific logic to fire when this occurs.
       * While the connection is down, the TreeCache will continue to have its state from before it lost
       * the connection and after the connection is restored, the TreeCache will emit normal child events
       * for all of the adds, deletes and updates that happened during the time that it was disconnected.
       * </p><p>
       * After reconnection, the cache will resynchronize its internal state with the server, then fire a
       * {@link #INITIALIZED} event.
       * </p>
       */
    TreeCacheEventType[TreeCacheEventType["CONNECTION_RECONNECTED"] = 4] = "CONNECTION_RECONNECTED";
    /**
       * Called when the connection has changed to {@link ConnectionState#LOST}
       * <p>
       * This is exposed so that users of the class can be notified of issues that *might* affect normal operation.
       * The TreeCache is written such that listeners are not expected to do anything special on this
       * event, except for those people who want to cause some application-specific logic to fire when this occurs.
       * While the connection is down, the TreeCache will continue to have its state from before it lost
       * the connection and after the connection is restored, the TreeCache will emit normal child events
       * for all of the adds, deletes and updates that happened during the time that it was disconnected.
       * </p>
       */
    TreeCacheEventType[TreeCacheEventType["CONNECTION_LOST"] = 5] = "CONNECTION_LOST";
    /**
       * Posted after the initial cache has been fully populated.
       * <p>
       * On startup, the cache synchronizes its internal
       * state with the server, publishing a series of {@link #NODE_ADDED} events as new nodes are discovered.  Once
       * the cachehas been fully synchronized, this {@link #INITIALIZED} this event is published.  All events
       * published after this event represent actual server-side mutations.
       * </p><p>
       * On reconnection, the cache will resynchronize its internal state with the server, and fire this event again
       * once its internal state is completely refreshed.
       * </p><p>
       * Note: because the initial population is inherently asynchronous, so it's possible to observe server-side changes
       * (such as a {@link #NODE_UPDATED}) prior to this event being published.
       * </p>
       */
    TreeCacheEventType[TreeCacheEventType["INITIALIZED"] = 6] = "INITIALIZED";
})(exports.TreeCacheEventType || (exports.TreeCacheEventType = {}));
/**
 * @private
 */
var TreeState;
(function (TreeState) {
    TreeState[TreeState["LATENT"] = 0] = "LATENT";
    TreeState[TreeState["STARTED"] = 1] = "STARTED";
    TreeState[TreeState["CLOSED"] = 2] = "CLOSED";
})(TreeState || (TreeState = {}));
/**
 * @private
 */
class TreeCacheImpl {
    // noinspection JSUnusedGlobalSymbols because it's necessary to satisfy TreeCacheView
    constructor(client, logger, path, cacheData, maxDepth, createParentNodes, selector, outstandingOps = 0) {
        this.client = client;
        this.logger = logger;
        this.cacheData = cacheData;
        this.maxDepth = maxDepth;
        this.createParentNodes = createParentNodes;
        this.selector = selector;
        this.outstandingOps = outstandingOps;
        // Have we published the {@link TreeCacheEventType#INITIALIZED} event yet?
        this.isInitialized = false;
        this._treeState = TreeState.LATENT;
        this.listeners = [];
        this.errorListeners = [];
        this.connectionStateListener = (_, newState) => {
            this.handleStateChange(newState);
        };
        this.root = new TreeNode(validatePath(path), null, this, this.logger);
        checkNotNull(this.client, `curator client cannot be null`);
    }
    treeState() {
        return this._treeState;
    }
    start() {
        if (this._treeState === TreeState.STARTED)
            throw new Error("cache already started");
        this._treeState = TreeState.STARTED;
        return new Promise((resolve, reject) => {
            const next = (err) => {
                if (err)
                    return reject(err);
                this.client.addConnectionStateListener(this.connectionStateListener);
                if (this.client.isConnected()) {
                    this.root.wasCreated();
                }
                resolve(this);
            };
            if (this.createParentNodes) {
                this.client.createContainers(this.root.path, next);
            }
            else {
                next(null);
            }
        });
    }
    close() {
        this.logger.debug("TreeCache#close() when tree state = ", this._treeState, " and have ", this.listeners.length, " listeners");
        if (this._treeState === TreeState.STARTED) {
            this._treeState = TreeState.CLOSED;
            this.client.removeConnectionStateListener(this.connectionStateListener);
            this.listeners.length = 0; // clear the array
            try {
                this.root.wasDeleted();
            }
            catch (e) {
                this.handleException(e);
            }
        }
    }
    getCurrentChildren(fullPath) {
        const node = this.find(fullPath);
        if (node === null || node.nodeState !== NodeState.LIVE) {
            return null;
        }
        const map = node.children;
        let result = new Map();
        if (map !== null) {
            for (let [key, childNode] of map) {
                const childData = childNode.childData;
                if (childData !== null && childNode.nodeState === NodeState.LIVE) {
                    result.set(key, childData);
                }
            }
        }
        return node.nodeState === NodeState.LIVE ? result : null;
    }
    getCurrentData(fullPath) {
        const node = this.find(fullPath);
        if (node === null || node.nodeState !== NodeState.LIVE) {
            return null;
        }
        const result = node.childData;
        return node.nodeState === NodeState.LIVE ? result : null;
    }
    addListener(listener) {
        this.listeners.push(listener);
        return this;
    }
    removeListener(listener) {
        const i = this.listeners.indexOf(listener);
        if (i > -1) {
            this.listeners.splice(i, 1);
        }
        return this;
    }
    addErrorListener(listener) {
        this.errorListeners.push(listener);
        return this;
    }
    removeErrorListener(listener) {
        const i = this.errorListeners.indexOf(listener);
        if (i > -1) {
            this.errorListeners.splice(i, 1);
        }
        return this;
    }
    handleException(e) {
        if (this.errorListeners.length === 0) {
            this.logger.error("", e);
        }
        else {
            this.errorListeners.forEach(listener => {
                try {
                    listener("", e);
                }
                catch (e) {
                    this.logger.error("Exception handling exception", e);
                }
            });
        }
    }
    publishEventType(type) {
        this.publishEvent(new TreeCacheEvent(type, null, new Date()));
    }
    publishEventTypeWithData(type, data) {
        this.publishEvent(new TreeCacheEvent(type, data, new Date()));
    }
    find(findPath) {
        const rootElements = split(this.root.path);
        const findElements = split(findPath);
        while (rootElements.length > 0) {
            if (findElements.length === 0) {
                // Target path shorter than root path
                return null;
            }
            const nextRoot = rootElements.shift();
            const nextFind = findElements.shift();
            if (nextFind !== nextRoot) {
                // Initial root path does not match
                return null;
            }
        }
        let current = this.root;
        while (findElements.length > 0) {
            const nextFind = findElements.shift();
            const map = current.children;
            if (map == null)
                return null;
            if (nextFind === undefined)
                return null;
            let nextCurrent = map.get(nextFind);
            if (nextCurrent === undefined)
                return null;
            current = nextCurrent;
        }
        return current;
    }
    handleStateChange(newState) {
        switch (newState) {
            case ConnectionState.SUSPENDED:
                this.publishEventType(exports.TreeCacheEventType.CONNECTION_SUSPENDED);
                break;
            case ConnectionState.LOST:
                this.isInitialized = false;
                this.publishEventType(exports.TreeCacheEventType.CONNECTION_LOST);
                break;
            case ConnectionState.CONNECTED:
                try {
                    this.root.wasCreated();
                }
                catch (e) {
                    this.handleException(e);
                }
                break;
            case ConnectionState.RECONNECTED:
                try {
                    this.root.wasReconnected();
                    this.publishEventType(exports.TreeCacheEventType.CONNECTION_RECONNECTED);
                }
                catch (e) {
                    this.handleException(e);
                }
                break;
            case ConnectionState.READ_ONLY:
        }
    }
    publishEvent(ev) {
        if (this._treeState !== TreeState.CLOSED) {
            const loggedEvent = Object.assign({}, ev, {
                _data: Object.assign({}, ev.data, { stat: "hidden" })
            });
            this.logger.debug("publishEvent", loggedEvent);
            process.nextTick(() => {
                this.listeners.forEach(listener => {
                    try {
                        listener(this.client, ev);
                    }
                    catch (e) {
                        this.handleException(e);
                    }
                });
            });
        }
    }
}

/**
 * @module node-zk-treecache
 */
/**  */
/**
 * @private
 */
class ConnectionStateManager {
    constructor(zkClient, curator) {
        this.zkClient = zkClient;
        this.numberOfConnectEvent = 0;
        this.listeners = [];
        zkClient.addListener("state", state => {
            if (state === nodeZookeeperClient.State.SYNC_CONNECTED ||
                state === nodeZookeeperClient.State.SASL_AUTHENTICATED) {
                this.numberOfConnectEvent++;
            }
            this.listeners.forEach(listener => {
                try {
                    listener(curator, this.zkStateToCS(state));
                }
                catch (e) {
                    // TODO Manage this error
                }
            });
        });
    }
    addConnectionStateListener(listener) {
        this.listeners.push(listener);
    }
    removeConnectionStateListener(listener) {
        const i = this.listeners.indexOf(listener);
        if (i > -1) {
            this.listeners.splice(i, 1);
        }
    }
    get connectionState() {
        return this.zkStateToCS(this.zkClient.getState());
    }
    get isConnected() {
        return isConnected(this.connectionState);
    }
    zkStateToCS(s) {
        switch (s) {
            case nodeZookeeperClient.State.DISCONNECTED:
                return ConnectionState.SUSPENDED;
            case nodeZookeeperClient.State.SYNC_CONNECTED:
                return this.numberOfConnectEvent > 1
                    ? ConnectionState.RECONNECTED
                    : ConnectionState.CONNECTED;
            case nodeZookeeperClient.State.AUTH_FAILED:
                return ConnectionState.LOST;
            case nodeZookeeperClient.State.CONNECTED_READ_ONLY:
                return ConnectionState.READ_ONLY;
            case nodeZookeeperClient.State.SASL_AUTHENTICATED:
                return this.numberOfConnectEvent > 1
                    ? ConnectionState.RECONNECTED
                    : ConnectionState.CONNECTED;
            case nodeZookeeperClient.State.EXPIRED:
                return ConnectionState.LOST;
            default:
                return ConnectionState.SUSPENDED;
        }
    }
}

/**
 * @module node-zk-treecache
 */
/**  */
function newCuratorFramework(zkClient) {
    const zk = zkClient;
    zk.connectionManager.setMaxListeners(100);
    return new CuratorFrameworkImpl(zk);
}
/**
 * @private
 */
class CuratorFrameworkImpl {
    constructor(zkClient) {
        this.zkClient = zkClient;
        this.connectionStateManager = new ConnectionStateManager(this.zkClient, this);
    }
    createContainers(path, cb) {
        mkdirs(this.zkClient, path, true, null, true, cb);
    }
    addConnectionStateListener(listener) {
        this.connectionStateManager.addConnectionStateListener(listener);
    }
    removeConnectionStateListener(listener) {
        this.connectionStateManager.removeConnectionStateListener(listener);
    }
    // if the ZK client is connected
    isConnected() {
        return this.connectionStateManager.isConnected;
    }
    getChildren() {
        return newGetChildrenBuilder(this);
    }
    getData() {
        return newGetDataBuilder(this);
    }
    checkExists() {
        return newExistsBuilder(this);
    }
}

/**
 * @module node-zk-treecache
 */
/**  */
const consoleLogger = (id, console) => ({
    id: Math.ceil(id * 1000),
    debug(msg, ...args) {
        console.log(`[${this.id}][debug]${msg}`, ...args);
    },
    info(msg, ...args) {
        console.log(`[${this.id}][info]${msg}`, ...args);
    },
    warn(msg, ...args) {
        console.log(`[${this.id}][warn]${msg}`, ...args);
    },
    error(msg, ...args) {
        console.log(`[${this.id}][error]${msg}`, ...args);
    }
});
class TreeCacheBuilder {
    constructor(client, path) {
        this.client = client;
        this.path = path;
        this.cacheData = true;
        this.maxDepth = Number.MAX_VALUE;
        this.createParentNodes = false;
        this.selector = new DefaultTreeCacheSelector();
        this.logger = consoleLogger(Math.random(), console);
    }
    build() {
        return new TreeCacheImpl(newCuratorFramework(this.client), this.logger, // TODO Expose this in the builder
        this.path, this.cacheData, this.maxDepth, this.createParentNodes, this.selector);
    }
    withLogger(logger) {
        this.logger = logger;
        return this;
    }
    withCacheData(cacheData) {
        this.cacheData = cacheData;
        return this;
    }
    withMaxDepth(maxDepth) {
        this.maxDepth = maxDepth;
        return this;
    }
    withCreateParentNodes(createParentNodes) {
        this.createParentNodes = createParentNodes;
        return this;
    }
    withSelector(selector) {
        this.selector = selector;
        return this;
    }
}

/**
 * Exported classes, this is what is available outside for users
 * TODO Tweak the documentation to reflect this property
 *
 * @module node-zk-treecache
 */
/** */
function treeCacheBuilder(client, path) {
    return new TreeCacheBuilder(client, path);
}

exports.TreeCacheBuilder = TreeCacheBuilder;
exports.treeCacheBuilder = treeCacheBuilder;
exports.TreeCacheEvent = TreeCacheEvent;

Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=node-zk-treecache.umd.js.map
