/// <reference types="node" />
import { CuratorEvent } from "./CuratorEvent";
import { CuratorFramework, CuratorFrameworkImpl } from "./CuratorFramework";
export declare enum ZooKeeperEventType {
    NODE_CREATED = 0,
    NODE_DELETED = 1,
    NODE_DATA_CHANGED = 2,
    NODE_CHILDREN_CHANGED = 3,
}
export interface ZooKeeperEvent {
    readonly type: ZooKeeperEventType;
    readonly name: string;
    readonly path: string;
    toString(): string;
}
export interface CuratorCallback {
    onCuratorEvent(client: CuratorFramework, event: CuratorEvent): void;
}
export interface ZooKeeperWatcher {
    onZooKeeperEvent(event: ZooKeeperEvent): void;
}
/**
 * @private
 * @param {CuratorFrameworkImpl} client
 * @returns {GetChildrenBuilder}
 */
export declare function newGetChildrenBuilder(client: CuratorFrameworkImpl): GetChildrenBuilder;
export interface GetChildrenBuilder {
    withWatcher(watcher: ZooKeeperWatcher): GetChildrenBuilder;
    withCallback(cb: CuratorCallback): GetChildrenBuilder;
    forPath(path: string): Promise<string[]>;
}
export declare function newGetDataBuilder(client: CuratorFrameworkImpl): GetDataBuilder;
export interface GetDataBuilder {
    withWatcher(watcher: ZooKeeperWatcher): GetDataBuilder;
    withCallback(cb: CuratorCallback): GetDataBuilder;
    forPath(path: string): Promise<Buffer | undefined>;
}
export declare function newExistsBuilder(client: CuratorFrameworkImpl): ExistsBuilder;
export interface ExistsBuilder {
    withWatcher(watcher: ZooKeeperWatcher): ExistsBuilder;
    withCallback(cb: CuratorCallback): ExistsBuilder;
    forPath(path: string): Promise<boolean>;
}
