/// <reference types="node" />
/**
 * @module node-zk-treecache
 */
/**  */
import { Stat } from "node-zookeeper-client";
import { ZooKeeperStat } from "./TreeCache";
/**
 * The kind of event emitted by the curator.
 * Please note that not all of them are implemented yet.
 */
export declare enum CuratorEventType {
    /**
     * Corresponds to {@link CuratorFramework#create()}
     */
    /**
     * Corresponds to {@link CuratorFramework#delete()}
     */
    /**
     * Corresponds to {@link CuratorFramework#checkExists()}
     */
    EXISTS = 0,
    /**
       * Corresponds to {@link CuratorFramework#getData()}
       */
    GET_DATA = 1,
    /**
       * Corresponds to {@link CuratorFramework#setData()}
       */
    /**
       * Corresponds to {@link CuratorFramework#getChildren()}
       */
    CHILDREN = 2,
}
export declare type CuratorEvent = GetDataCuratorEvent | GetChildrenCuratorEvent | ExistsCuratorEvent;
export declare enum ZooKeeperCode {
    OK = 0,
    SYSTEMERROR = -1,
    RUNTIMEINCONSISTENCY = -2,
    DATAINCONSISTENCY = -3,
    CONNECTIONLOSS = -4,
    MARSHALLINGERROR = -5,
    UNIMPLEMENTED = -6,
    OPERATIONTIMEOUT = -7,
    BADARGUMENTS = -8,
    NEWCONFIGNOQUORUM = -13,
    RECONFIGINPROGRESS = -14,
    UNKNOWNSESSION = -12,
    APIERROR = -100,
    NONODE = -101,
    NOAUTH = -102,
    BADVERSION = -103,
    NOCHILDRENFOREPHEMERALS = -108,
    NODEEXISTS = -110,
    NOTEMPTY = -111,
    SESSIONEXPIRED = -112,
    INVALIDCALLBACK = -113,
    INVALIDACL = -114,
    AUTHFAILED = -115,
    SESSIONMOVED = -118,
    NOTREADONLY = -119,
    EPHEMERALONLOCALSESSION = -120,
    NOWATCHER = -121,
    RECONFIGDISABLED = -123,
}
export interface GetDataCuratorEvent {
    readonly type: CuratorEventType.GET_DATA;
    readonly path: string;
    readonly stat: ZooKeeperStat | undefined;
    readonly data: Buffer | undefined;
    readonly returnCode: ZooKeeperCode;
}
export interface GetChildrenCuratorEvent {
    readonly type: CuratorEventType.CHILDREN;
    readonly path: string;
    readonly stat: ZooKeeperStat | undefined;
    readonly children: string[] | undefined;
    readonly returnCode: ZooKeeperCode;
}
export interface ExistsCuratorEvent {
    readonly type: CuratorEventType.EXISTS;
    readonly path: string;
    readonly stat: ZooKeeperStat | null;
    readonly returnCode: ZooKeeperCode;
}
/**
 *
 * @private
 * @param path
 * @param stat
 * @param data
 * @param returnCode
 */
export declare function createGetDataCuratorEvent(path: string, stat: Stat | undefined, data: Buffer | undefined, returnCode: ZooKeeperCode): GetDataCuratorEvent;
/**
 *
 * @private
 * @param path
 * @param stat
 * @param children
 * @param returnCode
 */
export declare function createGetChildrenCuratorEvent(path: string, stat: Stat | undefined, children: string[] | undefined, returnCode: ZooKeeperCode): GetChildrenCuratorEvent;
export declare function createExistsCuratorEvent(path: string, stat: Stat | null, returnCode: ZooKeeperCode): ExistsCuratorEvent;
