import { Client, ACL, Exception } from "node-zookeeper-client";
/**
 * @private
 * @param path the path to split
 */
export declare function split(path: string): string[];
export interface ACLProvider {
    getDefaultAcl(): ACL[] | null;
    getAclForPath(path: string): ACL[] | null;
}
/**
 * Make sure all the nodes in the path are created. NOTE: Unlike File.mkdirs(), Zookeeper doesn't distinguish
 * between directories and files. So, every node in the path is created. The data for each node is an empty blob
 *
 * @private
 * @param zkClient     the client
 * @param path         path to ensure
 * @param makeLastNode if true, all nodes are created. If false, only the parent nodes are created
 * @param aclProvider  if not null, the ACL provider to use when creating parent nodes
 * @param asContainers if true, nodes are created as {@link CreateMode#CONTAINER} (need ZK > 3.5)
 * @param cb           the callback to call after having created the path
 */
export declare function mkdirs(zkClient: Client, path: string, makeLastNode: boolean, aclProvider: ACLProvider | null, asContainers: boolean, cb?: (err: Error | Exception, p: string) => void): void;
/**
 * Given a full path, return the node name. i.e. "/one/two/three" will return "three"
 *
 * @private
 * @param path the path
 * @return the node
 */
export declare function getNodeFromPath(path: string): string;
/**
 * Given a parent path and a child node, create a combined full path
 *
 * @param parent the parent
 * @param child  the child
 * @return full path
 */
export declare function makePath(parent: string, child: string): string;
